/*
 *  Simple OpenCL demo program
 *
 *  Copyright (C) 2009  Clifford Wolf <clifford@clifford.at>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	Revision history:
 *  07 Jul 2015, Vladimir Smirnov <kkatarn@sleepgate.ru>:
 *	ANSIfication and restructurization. To compile:
 *		cc cldemo.c -pedantic -ansi -Wall -lOpenCL -o cldemo
 *	where cc can be either gcc or clang.
 */

#include <CL/cl.h>
#include <stdio.h>

#define	BUFLEN			10240
#define	MAX_PLATFORMS	128
#define	MAX_DEVS		128

#define VECTOR_LEN		16


/*
 * CL_WARNING_ON_FAILURE
 * This macro prints warning if retval != CL_SUCCESS
 * Argument is referenced exactly once.
 * TODO: print descriptive OpenCL error string.
 */
#define CL_WOF(retval) \
	do { \
		if ((retval) == CL_SUCCESS) \
			break; \
		fprintf(stderr, "OpenCL Error at line %d\n", __LINE__); \
   } while (0)

/*
 * CL_RETURN_ON_FAILURE
 * Same as above and issues 'return 1'.
 * Argument is referenced exactly once.
 */
#define CL_ROF(retval) \
	do { \
		if ((retval) == CL_SUCCESS) \
			break; \
		fprintf(stderr, "OpenCL Error at line %d\n", __LINE__); \
		return 1; \
   } while (0)

/*
 * create_cb()
 * Callback for clCreateContext()
 */
static void
create_cb(const char *errinfo, const void *private_info, size_t cb, void *user_data)
{
	fprintf(stderr, "clCreateContext() error: %s\n", errinfo);
}

/*
 * build_cb()
 * Callback for clBuildProgram()
 * It does not used because with such callback clBuildProgram() may not block.
 * This is not what we want.
 *
void build_cb(cl_program program, void *user_data)
{
	fprintf(stderr, "Build of the program finished.\n");
}
*/

/*
 * run_kernel()
 * For the given platform ID, number of devices and list of device IDs
 * compiles and runs simple kernel that multiplies scalar by vector.
 * Returns 0 on success.
 */
static int
run_kernel(cl_platform_id platform,cl_uint num_devs,const cl_device_id *devs)
{
	cl_int retval;
	cl_program program;
	cl_context context;
	int i, factor = 3; /* scale to multiply vector to */
	cl_mem input_buffer;
	cl_mem output_buffer;
	cl_kernel kernel;
	cl_command_queue queue;
	cl_event kernel_completion;
	size_t global_work_size[1] = { VECTOR_LEN };
	const char *program_source[] = {
		"__kernel void simple_demo(__global int *src, __global int *dst, int factor)\n",
		"{\n",
		"	int i = get_global_id(0);\n",
		"	dst[i] = src[i] * factor;\n",
		"}\n"
	};
	cl_context_properties create_props[] = { CL_CONTEXT_PLATFORM, 0, 0 };
	create_props[1] = (cl_context_properties) platform;
	printf("\tBuilding test OpenCL program...\n");
	context = clCreateContext(create_props, num_devs, devs, &create_cb, NULL, &retval);
	CL_ROF(retval);
	program = clCreateProgramWithSource(context, sizeof(program_source)/sizeof(*program_source), program_source, NULL, &retval);
	CL_ROF(retval);
	if (clBuildProgram(program, num_devs, devs, "", /*&build_cb*/ NULL, NULL) != CL_SUCCESS) {
		char buffer[BUFLEN];
		/* FIXME: is this correct? */
		clGetProgramBuildInfo(program, devs[0], CL_PROGRAM_BUILD_LOG, sizeof(buffer), buffer, NULL);
		fprintf(stderr, "\t\tCL Compilation failed:\n%s", buffer);
		return 1;
	}
	/* CL_WOF(clUnloadCompiler()); this call is deprecated */
	input_buffer = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(int)*VECTOR_LEN, NULL, &retval);
	CL_ROF(retval);
	output_buffer = clCreateBuffer(context, CL_MEM_WRITE_ONLY, sizeof(int)*VECTOR_LEN, NULL, &retval);
	CL_ROF(retval);
	kernel = clCreateKernel(program, "simple_demo", &retval);
	CL_ROF(retval);
	CL_ROF(clSetKernelArg(kernel, 0, sizeof(input_buffer), &input_buffer));
	CL_ROF(clSetKernelArg(kernel, 1, sizeof(output_buffer), &output_buffer));
	CL_ROF(clSetKernelArg(kernel, 2, sizeof(factor), &factor));
	queue = clCreateCommandQueue(context, devs[0], 0, &retval); /* FIXME: is this correct? */
	CL_ROF(retval);
	for (i=0; i<VECTOR_LEN; i++) {
		CL_ROF(clEnqueueWriteBuffer(queue, input_buffer, CL_TRUE, i*sizeof(int), sizeof(int), &i, 0, NULL, NULL));
	}

	printf("\tStarting execution of the kernel on %u device(s)...\n", num_devs);
	CL_ROF(clEnqueueNDRangeKernel(queue, kernel, 1, NULL, global_work_size, NULL, 0, NULL, &kernel_completion));
	CL_ROF(clWaitForEvents(1, &kernel_completion));
	printf("\tExecution of the kernel complete. Result of %d*(0,1,2,...):\n\t\t\t", factor);

	CL_ROF(clReleaseEvent(kernel_completion));
	for (i=0; i<VECTOR_LEN; i++) {
		int data;
		CL_ROF(clEnqueueReadBuffer(queue, output_buffer, CL_TRUE, i*sizeof(int), sizeof(int), &data, 0, NULL, NULL));
		printf(" %d", data);
	}
	printf("\n");
	CL_WOF(clReleaseMemObject(input_buffer));
	CL_WOF(clReleaseMemObject(output_buffer));
	CL_WOF(clReleaseKernel(kernel));
	CL_WOF(clReleaseProgram(program));
	CL_WOF(clReleaseContext(context));
	return 0;
}

int
main(int argc, char **argv)
{
	int i, j;
	char buffer[BUFLEN];
	cl_uint buf_uint;
	cl_ulong buf_ulong;
	cl_platform_id platforms[MAX_PLATFORMS];
	cl_uint platforms_n = 0;
	cl_device_id devices[MAX_DEVS];
	cl_uint devices_n = 0;
	CL_ROF(clGetPlatformIDs(MAX_PLATFORMS, platforms, &platforms_n));
	printf("=== %d OpenCL platform(s) found: ===\n", platforms_n);
	for (i=0; i<platforms_n; i++)
	{
		printf("\t-- %d --\n", i);
		CL_ROF(clGetPlatformInfo(platforms[i], CL_PLATFORM_PROFILE, BUFLEN, buffer, NULL));
		printf("\tPROFILE = %s\n", buffer);
		CL_ROF(clGetPlatformInfo(platforms[i], CL_PLATFORM_VERSION, BUFLEN, buffer, NULL));
		printf("\tVERSION = %s\n", buffer);
		CL_ROF(clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, BUFLEN, buffer, NULL));
		printf("\tNAME = %s\n", buffer);
		CL_ROF(clGetPlatformInfo(platforms[i], CL_PLATFORM_EXTENSIONS, BUFLEN, buffer, NULL));
		printf("\tEXTENSIONS = %s\n", buffer);
		CL_ROF(clGetPlatformInfo(platforms[i], CL_PLATFORM_VENDOR, BUFLEN, buffer, NULL));
		printf("\tVENDOR = %s\n", buffer);
		/* Vendor name is now in 'buffer' */
		devices_n = 0;
		if((clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_GPU | CL_DEVICE_TYPE_CPU, MAX_DEVS, devices, &devices_n) != CL_SUCCESS) || !devices_n)
		{
			printf("\t=== no OpenCL-capable devices found for the platform \"%s\".\n", buffer);
			continue;
		}
		printf("\t=== %d OpenCL-capable device(s) found for the platform \"%s\".\n", devices_n, buffer);
		for (j=0; j<devices_n; j++)
		{
			printf("\t\tPlatform \"%s\", device #%d --\n", buffer, j);
			CL_ROF(clGetDeviceInfo(devices[j], CL_DEVICE_NAME, sizeof(buffer), buffer, NULL));
			printf("\t\tDEVICE_NAME = %s\n", buffer);
			CL_ROF(clGetDeviceInfo(devices[j], CL_DEVICE_VENDOR, sizeof(buffer), buffer, NULL));
			printf("\t\tDEVICE_VENDOR = %s\n", buffer);
			CL_ROF(clGetDeviceInfo(devices[j], CL_DEVICE_VERSION, sizeof(buffer), buffer, NULL));
			printf("\t\tDEVICE_VERSION = %s\n", buffer);
			CL_ROF(clGetDeviceInfo(devices[j], CL_DRIVER_VERSION, sizeof(buffer), buffer, NULL));
			printf("\t\tDRIVER_VERSION = %s\n", buffer);
			CL_ROF(clGetDeviceInfo(devices[j], CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(buf_uint), &buf_uint, NULL));
			printf("\t\tDEVICE_MAX_COMPUTE_UNITS = %u\n", (unsigned int)buf_uint);
			CL_ROF(clGetDeviceInfo(devices[j], CL_DEVICE_MAX_CLOCK_FREQUENCY, sizeof(buf_uint), &buf_uint, NULL));
			printf("\t\tDEVICE_MAX_CLOCK_FREQUENCY = %u\n", (unsigned int)buf_uint);
			CL_ROF(clGetDeviceInfo(devices[j], CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(buf_ulong), &buf_ulong, NULL));
			buf_ulong = buf_ulong >> 20;
			printf("\t\tDEVICE_GLOBAL_MEM_SIZE = %lu MB\n", (unsigned long)buf_ulong);
		}
		run_kernel(platforms[i], devices_n, devices);
	}
	return 0;
}
